import sys
import logging
import subprocess
from enum import Enum
from typing import Optional, Union, List

class Resultado(Enum):
	''' Resultado de ejecución de un subproceso '''

	EXITO = 0
	ERROR = 1

	@classmethod
	def _missing_(cls, _):
		return cls.ERROR

	def __bool__(self):
		return self == Resultado.EXITO

class LogHandle:
	''' Permite redirigir la salida estándar y de error a un logger '''

	def __init__(self, level: int):
		self.level = level
		self.message = ''

	def write(self, message: str):
		self.message = self.message + message

		while '\n' in self.message:
			index = self.message.index('\n')
			logging.log(self.level, self.message[:index])
			self.message = self.message[(index + 1):]

	def flush(self):
		pass

class TerminalColor(Enum):
	''' Colores para la terminal '''

	GREY     = '\x1b[38;5;248m'
	BLUE     = '\x1b[38;5;39m'
	YELLOW   = '\x1b[38;5;226m'
	RED      = '\x1b[38;5;196m'
	BOLD_RED = '\x1b[31;1m'
	RESET    = '\x1b[0m'

	def wrap(self, text: str):
		return self.value + text + TerminalColor.RESET.value

class ConsoleFormatter(logging.Formatter):
	''' Formatea los logs para que se vean bien en la terminal '''

	def __init__(self, format: str):
		super().__init__()
		self.fmt = format
		self.formats = \
		{
			logging.DEBUG:    TerminalColor.GREY.wrap(self.fmt),
			logging.INFO:     TerminalColor.BLUE.wrap(self.fmt),
			logging.WARNING:  TerminalColor.YELLOW.wrap(self.fmt),
			logging.ERROR:    TerminalColor.RED.wrap(self.fmt),
			logging.CRITICAL: TerminalColor.BOLD_RED.wrap(self.fmt)
		}

	def format(self, record: logging.LogRecord):
		format = self.formats.get(record.levelno)
		formatter = logging.Formatter(format)
		return formatter.format(record)

def configurar_logs(archivo: Optional[str] = None):
	''' Configura los logs para redirigirlos a la salida estándar y a un archivo '''

	# Formatos para los logs
	FORMATO_TERMINAL = '%(message)s'
	FORMATO_ARCHIVO  = '[%(asctime)-19s] [%(levelname)-8s]: %(message)s'

	# Se redirigen los logs a la salida estándar y se configura el formato
	manejadores = [logging.StreamHandler(sys.stdout)]
	manejadores[0].setFormatter(ConsoleFormatter(FORMATO_TERMINAL))

	# Si se especificó un archivo, se redirigen los logs a él también
	if archivo:
		manejadores.append(logging.FileHandler(archivo))

	# Se configura el logger
	logging.basicConfig \
	(
		datefmt= '%d/%m/%Y %H:%M:%S',
		format= FORMATO_ARCHIVO,
		handlers= manejadores,
		level= logging.DEBUG
	)

	# Se redirigen las salidas estándar y de error al logger
	sys.stdout = LogHandle(logging.DEBUG)
	sys.stderr = LogHandle(logging.ERROR)

	# Se desactivan los logs de la librería 'requests'
	logging.getLogger('urllib3').setLevel(logging.WARNING)

def subproceso \
(
	comandos: Union[str, List[str]],
	mostrar_stdout: bool = False,
	mostrar_stderr: bool = True,
	regresar_stdout: bool = False,
	redirigir_stderr: bool = False
):
	''' Ejecuta un subproceso y redirige su salida a los logs '''

	sub = subprocess.run(comandos, capture_output= True, shell= isinstance(comandos, str))

	# Se imprimen los logs de la salida estándar
	if mostrar_stdout:
		for line in sub.stdout.decode('utf-8').split('\n'):
			if len(line) > 0:
				logging.debug(line)

	# Se imprimen los logs de la salida de error
	if mostrar_stderr:
		for line in sub.stderr.decode('utf-8').split('\n'):
			if len(line) > 0:
				if not redirigir_stderr:
					logging.error(line)
				else:
					logging.info(line)

	# Se retorna la salida estándar si se solicita
	if regresar_stdout:
		return sub.stdout.decode('utf-8') if sub.returncode == Resultado.EXITO.value else ''

	# Si la salida estándar contiene el mensaje 'error_code' se retorna un error
	if 'error_code' in sub.stdout.decode('utf-8'):
		return Resultado.ERROR

	return Resultado(sub.returncode)

def info(mensaje: str):
	''' Muestra un mensaje de información '''

	logging.info(mensaje)

def error(mensaje: str, salir= True):
	''' Muestra un mensaje de error y opcionalmente termina el programa '''

	logging.error(mensaje)

	# Se termina el programa si se solicita
	if salir:
		sys.exit(Resultado.ERROR.value)